#!/usr/bin/perl -w
#
#  $Id: expandDomains.pl,v 1.2 2007/10/16 23:20:20 whuang Exp $
#  fastHmm/fastBlast Alignment Tools
#  http://microbesonline.org/fasthmm (fasthmm@microbesonline.org)
#
#  Expand blast hits to domain families
#
#  Copyright (C) 2007 The Regents of the University of California
#  All rights reserved.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#  Disclaimer
#
#  NEITHER THE UNITED STATES NOR THE UNITED STATES DEPARTMENT OF ENERGY,
#  NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED,
#  OR ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY,
#  COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
#  OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
#  PRIVATELY OWNED RIGHTS.
#

use strict;

use lib exists( $ENV{FASTHMM_DIR} ) ?
		"$ENV{FASTHMM_DIR}/lib" :
		"./lib";
use Args;
use File;
use Util;
use Vector;


#
# USAGE
#
my $usage =
"Usage:
  expandDomains.pl <options>

Parameters:
  -domains  <inFile>	Specify reduced domains file
  -clusters <blastDb>	Specify corresponding extracted subsequence FASTA file

Optional Parameters:
  -o        <numAA>	Specify minimum overlap for expansion in aa
			   Default: 30aa
  -c        <%-cov>	Specify minimum %-coverage for expansion [0-1]
			   Default: 0.5
  -a			Both -o and -c must be true in order for expansion
			   Default: only one of -o or -c must be true
  -output   <file>	Specify file to which results will be written
			   Default: output is generated to stdout
  -debug		Enable additional debugging output
  -sorted		Input files are both sorted by sequence id.  This will
			considerably reduce the memory footprint.
  -lowmem		Slower but will prevent caching of unused cluster entries

";

#
# Parse Command Line Options
#
my ($opts, $nonOpts) = Args::getArgs( 
	"+domains:|+clusters:|output:|debug|o:|c:|a|sorted|lowmem",
	@ARGV, -1, $usage );

die "Error: Specified input file '", $opts->{domains}, "' does not exist!\n"
	if ( !(-e $opts->{domains}) );

die "Error: Specified clusters file '", $opts->{clusters}, "' does not exist!\n"
	if ( !(-e $opts->{clusters}) );

$opts->{o} = 30
	if ( !exists( $opts->{o} ) );

$opts->{c} = 0.5
	if ( !exists( $opts->{c} ) );

expandDomains( $opts );

exit(0);

sub buildIdInclusionHash
{
	my $file = shift;
	my %hash = ();
	local *IN;

	open( IN, "<$file" );
	while( <IN> )
	{
		chomp;
		my ( undef, $gene ) = split( /\t/, $_, 4 );
		my ( $locusId ) = $gene =~ /^(.+?)\.\d+\.\d+$/;
		$hash{$locusId} = 1;
	}
	close( IN );

	return \%hash;
}

sub expandDomains
{
	my $opts = shift;
	local *DOMAINS;
	local *CLUSTERS;
	local *OUT;

	# open output file
	if ( exists( $opts->{output} ) )
	{
		open( OUT, ">$opts->{output}" );
	} else {
		open( OUT, ">&=1" );
	}

	# build low-mem inclusion hash?
	my ( $hash, $exclusion );
	if ( exists( $opts->{lowmem} ) )
	{
		$hash = buildIdInclusionHash( $opts->{domains} );
		$exclusion = 0;
	} else {
		$hash = {};
		$exclusion = 1;
	}

	# iterate over domains file
	my $fId = File::openFileIterCache( \*CLUSTERS, $opts->{clusters},
			'^(\d+)',
			$hash,
			$exclusion,
			!exists( $opts->{sorted} ),
			exists( $opts->{debug} ) );
	open( DOMAINS, "<$opts->{domains}" );

	my $curId = undef;
	my $curGroup = undef;
	my $numDomains = 0;
	my $numExpanded = 0;

	while( <DOMAINS> )
	{
		chomp;
		my ( $query, $subject, $beg, $end ) = split(/\t/, $_, 4);
		my ( $seqId ) = $subject =~ /^(.+?)\.\d+\.\d+$/;
		if ( !defined($curId) || ($curId ne $seqId) )
		{
			$curId = $seqId;
			$curGroup = File::getFileIterIdGroupCache( $fId, $seqId );
			if ( !defined($curGroup) || ( scalar(@{$curGroup}) < 1 ) )
			{
				print STDERR "no cluster group for id '$seqId'\n"
					if ( exists( $opts->{debug} ) );
			}
		}

		my ( $locusId, $regBeg, $regEnd ) = split(/\./, $subject, 3);
		my $hitBeg = $regBeg + $beg - 1;
		my $hitEnd = $regBeg + $end - 1;


		# output exact hit
		print OUT join( "\t", $query, $locusId, $hitBeg, $hitEnd, 'exact' ), "\n";
		$numDomains++;
		$numExpanded++;

		# check for expansion
		if ( defined($curGroup) && ( scalar( @{$curGroup} ) > 0 ) )
		{
			foreach my $line ( @{$curGroup} )
			{
				my ( $exemplar, $exact, @others ) = split( /\t/, $line );
				# nothing to expand to anyway
				next if ( scalar(@others) < 1 );

				my ( $eLocusId, $eBeg, $eEnd ) = $exemplar =~ /^(.+?)\.(\d+)\.(\d+)$/;
				die "attempting to use cluster group $eLocusId with hit $subject"
					if ( $eLocusId ne $locusId );

				my $overlap = ( ( $hitEnd < $eEnd ) ? $hitEnd : $eEnd ) -
						( ( $hitBeg > $eBeg ) ? $hitBeg : $eBeg ) + 1;
				next if ( $overlap < 1 );

				my $pctCover = $overlap * 1.0 / ( $eEnd - $eBeg + 1 );
				my $expand = exists( $opts->{a} ) ?
						( ( $pctCover >= $opts->{c} ) && ( $overlap >= $opts->{o} ) ) :
						( ( $pctCover >= $opts->{c} ) || ( $overlap >= $opts->{o} ) );

				#print STDERR "$locusId:[$hitBeg,$hitEnd] <=> $eLocusId:[$eBeg,$eEnd] :: overlap:[$overlap aa], pctCover:[$pctCover], expand:[$expand]\n"
				#	if ( exists( $opts->{debug} ) );

				if ( $expand )
				{
					foreach my $gene ( @others )
					{
						my ( $cLocusId, $cBeg, $cEnd ) = $gene =~ /^(.+?)\.(\d+)\.(\d+)$/;
						print OUT join( "\t", $query, $cLocusId, $cBeg, $cEnd, 'approx' ), "\n";
						$numExpanded++;
					}
				}
			}
		}
	}

	close( DOMAINS );
	File::closeFileIterCache( $fId );

	close( OUT );

	print STDERR "Expanded $numDomains domains to $numExpanded domains\n";
}
