#!/usr/bin/perl -w
#
#  $Id: fastBlast.pl,v 1.4 2007/10/16 23:20:20 whuang Exp $
#  fastHmm/fastBlast Alignment Tools
#  http://microbesonline.org/fasthmm (fasthmm@microbesonline.org)
#
#  fastBlast Alignment Tool
#
#  Copyright (C) 2007 The Regents of the University of California
#  All rights reserved.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#  Disclaimer
#
#  NEITHER THE UNITED STATES NOR THE UNITED STATES DEPARTMENT OF ENERGY,
#  NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED,
#  OR ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY,
#  COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
#  OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
#  PRIVATELY OWNED RIGHTS.
#

use strict;
use lib exists( $ENV{FASTHMM_DIR} ) ?
		"$ENV{FASTHMM_DIR}/lib" :
		"./lib";
use Args;
use File;
use Util;

#
# USAGE
#
my $usage =
"Usage:
  fastBlast.pl <options>

Parameters:
  -i <dir>	fastHmm output directory containing result.*.hmmhits
  -f <fasta>	Input fasta file supplied to fastHmm

Optional Parameters:
  -c <confFile>	Configuration parameters file for fastBlast
  -o <outDir>	Directory in which to generate output
		   (Default: Use current working directory)
  -d <dbDir>	Specify database base directory
		   (Default: \$ENV{FASTHMM_DIR}/db)
  -D		Debugging mode; do not remove intermediate files
  -q		Quiet execution; disable status updates to stdout
  -j <procs>	Parallelize analysis of targets on a single host
  -t <tmpdir>	Specify base temporary directory

Distributed Computation Parameters:
  -b <size>	Specify number of query sequences per BLAST job;
		    Pre-processing will run on this host with -j parallel
		    processes and two command lists will be created, one
		    containing the BLAST jobs and one containing post-
		    processing which can only be run after all BLAST jobs
		    have completed.

";


#
# FastBLAST has two operating modes: 1) single-machine operation (supports
# one or more concurrent threads) and 2) distributed computation.  In
# single-machine mode, FastBLAST will run the analysis from start to
# completion using the specified number of threads (-j) when appropriate.
# In distributed computation mode, FastBLAST divides the tasks into three
# stages -- pre-processing, blast, post-processing.  Jobs from different
# stages may not be run concurrently.  You must submit the stages in order
# and wait for their completion before issuing the jobs to the subsequent
# stage.  FastBLAST will not run any of the jobs in distributed computation
# mode and instead will generate three command lists corresponding to the
# three stages.  All of the commands in each stage job list can be run in
# parallel.  Commands listed on the same line must be run in sequence on
# the same compute node.  For distributed computation mode, it is assumed
# all compute nodes write to a shared folder that is accessible by all
# machines.  FastBLAST does not provide a mechanism to move files between
# compute nodes.  To use distributed computation mode, simply specify
# the -b parameter.
#

#
# Parse Command Line Options
#
my ($opts, $nonOpts) = Args::getArgs( 
	"+i:|+f:|d:|o:|D|q|j:|b:|m:|t:|c:",
	@ARGV, -1, $usage );

$opts->{cmdline} = "$0 " . join(" ", @ARGV);

$opts->{c} = ( exists( $ENV{FASTHMM_DIR} ) ? $ENV{FASTHMM_DIR} : "." ) . "/conf/fastBlast.conf"
	if ( !exists( $opts->{c} ) );

my $conf = File::loadConfMacros( $opts->{c} );

$opts->{o} = "."
	if ( !exists( $opts->{o} ) );

$opts->{j} = $conf->{NUMTHREADS}
	if ( !exists( $opts->{j} ) );

$opts->{t} = $conf->{TMPDIR}
	if ( !exists( $opts->{t} ) );

die "Error: Specified fastHmm results directory '", $opts->{i}, "' does not exist!\n"
	if ( !(-d $opts->{i}) );

die "Error: Specified FASTA file '", $opts->{f}, "' does not exist!\n"
	if ( !(-e $opts->{f}) );

$opts->{d} = ( exists( $ENV{FASTHMM_DIR} ) ? $ENV{FASTHMM_DIR} : "." ) . "/db"
	if ( !exists( $opts->{d} ) );

die "Error: Specified output directory '", $opts->{o}, "' does not exist!\n"
	if ( !(-d $opts->{o}) );

$opts->{conf} = $conf;

if ( exists( $opts->{m} ) )
{
	if ( $opts->{m} =~ /^(?:pre|blast|post|cmdlist)$/i )
	{
		fastBlast( $opts, $opts->{m} );
	} else {
		die "Error: Unknown operating mode '$opts->{m}'... please try again!\n";
	}
} elsif ( !exists( $opts->{b} ) )
{
	foreach my $mode ( qw/pre blast post/ )
	{
		fastBlast( $opts, $mode );
	}
} else
{
	fastBlast( $opts, "cmdlist" );
}

exit(0);


sub fastBlast
{
	my $opts = shift;
	my $mode = shift;

	$mode = lc($mode);

	my $baseDir = exists( $ENV{FASTHMM_DIR} ) ?
				$ENV{FASTHMM_DIR} :
				".";
	$opts->{baseDir} = $baseDir;

	$ENV{'BLASTMAT'} = $baseDir . "/matrix";

	if ( $mode eq 'pre' )
	{
		fastBlast_pre( $opts, $baseDir );
	} elsif ( $mode eq 'blast' )
	{
		fastBlast_blast( $opts, $baseDir );
	} elsif ( $mode eq 'post' )
	{
		fastBlast_post( $opts, $baseDir );
	} elsif ( $mode eq 'cmdlist' )
	{
		fastBlast_pre( $opts, $baseDir );
		fastBlast_cmdlist( $opts, $baseDir );
	} else
	{
		die "Error: Unknown operating mode '$opts->{m}'... please try again!\n";
	}
}

sub fastBlast_cmdlist
{
	my $opts = shift;
	my $baseDir = shift;
	my $conf = $opts->{conf};

	# generate blast cmdlist
	dprint( "generating blast cmdlist '$opts->{o}/cmdlist.blast'\n", $opts );
	my $fileList = "$opts->{o}/file.list";
	my $blastCmdList = "$opts->{o}/cmdlist.blast";
	local *FL;
	local *OUT;

	open( FL, "<$fileList" );
	open( OUT, ">$blastCmdList" );
	while(<FL>)
	{
		chomp;
		my $inFile = $_;

		my ( $partNum ) = $inFile =~ /\.(\d+)$/;
		if ( defined($partNum) )
		{
			$partNum = "." . $partNum;
		} else {
			$partNum = "";
		}
		my $cmd = "$baseDir/bin/blastall -p blastp -a $opts->{j} -m 8 -i $inFile -d $opts->{o}/combined.faa -o $opts->{o}/blastp.out$partNum";
		my $optParams = Util::configParams( $conf, "BLASTP" );
		$cmd .= " $optParams"
			if ( length($optParams) > 0 );
		print OUT $cmd, "\n";
	}
	close(OUT);
	close(FL);

	# generate post cmdlist
	dprint( "generating post-processing cmdlist '$opts->{o}/cmdlist.post'\n", $opts );
	open(OUT, ">$opts->{o}/cmdlist.post");
	print OUT $opts->{cmdline} . " -m post\n";
	close(OUT);
}

sub fastBlast_post
{
	my $opts = shift;
	my $baseDir = shift;
	my $conf = $opts->{conf};

	my $files = File::getFilteredFileList( $opts->{o}, [ '^blastp\.out(?:\.\d+)?$' ] );
	dprint( "found " . scalar(@{$files}) . " blastp output files\n", $opts );
	if ( scalar(@{$files}) > 1 )
	{
		dprint( "merging blastp output files\n", $opts );
		concatenateFiles( "$opts->{o}/blastp.out", @{$files} );
		if ( !exists( $opts->{D} ) )
		{
			foreach my $file ( @{$files} )
			{
				unlink( $file );
			}
		}
	}

	# all blastp results in blastp.out

	dprint( "extracting domains from blastp output\n", $opts );
	local *IN;
	local *OUT;
	open(IN, "<$opts->{o}/blastp.out");
	open(OUT, "| $baseDir/bin/sort -nk 2 >$opts->{o}/blastp.domains.sorted");
	while(<IN>)
	{
		chomp;
		my @data = split(/\t/, $_);
		if ( scalar(@data) >= 11 )
		{
			print OUT join("\t", $data[0], $data[1], $data[8], $data[9]), "\n";
		}
	}
	close(OUT);
	close(IN);
	# unlink( "$opts->{o}/blastp.out", "$opts->{o}/combined.faa", "$opts->{o}/combined.faa.phr", "$opts->{o}/combined.faa.pin", "$opts->{o}/combined.faa.psq" );

	# now run expandDomains
	dprint( "expanding blast domains\n", $opts );
	my $cmd = "$baseDir/bin/expandDomains.pl -clusters $opts->{o}/combined.clstr.parsed.sorted -domains $opts->{o}/blastp.domains.sorted -sorted";
	my $optParams = Util::configParams( $conf, "EXPANDDOMAINS" );
	$cmd .= " $optParams"
		if ( length($optParams) > 0 );
	$cmd .= " >$opts->{o}/blastp.domains.expanded";
	system( $cmd );
	# unlink( "$opts->{o}/combined.clstr.parsed", "$opts->{o}/blastp.domains" );
}

sub fastBlast_blast
{
	my $opts = shift;
	my $baseDir = shift;
	my $conf = $opts->{conf};

	my $fileList = "$opts->{o}/file.list";
	local *FL;

	open( FL, "<$fileList" );
	while(<FL>)
	{
		chomp;
		my $inFile = $_;

		my ( $partNum ) = $inFile =~ /\.(\d+)$/;
		if ( defined($partNum) )
		{
			$partNum = "." . $partNum;
		} else {
			$partNum = "";
		}
		dprint( "running blast on $inFile ...\n", $opts );
		my $cmd = "$baseDir/bin/blastall -p blastp -a $opts->{j} -m 8 -i $inFile -d $opts->{o}/combined.faa -o $opts->{o}/blastp.out$partNum";
		my $optParams = Util::configParams( $conf, "BLASTP" );
		$cmd .= " $optParams"
			if ( length($optParams) > 0 );
		system( $cmd );
	}
	close(FL);
}

sub fastBlast_pre
{
	my $opts = shift;
	my $baseDir = shift;
	my $conf = $opts->{conf};

	my $files = File::getFilteredFileList( $opts->{i}, [ '^result\..+\.hmmhits$' ] );
	dprint( "concatenating " . scalar(@{$files}) . " hmmhits files; mapping pfam to clan ...\n", $opts );
	my $allDomainsFile = concatenateFastHmmDomains( $files, $opts );
	dprint( "wrote $allDomainsFile\n", $opts );

	# reduce domains
	my $outPrefix = $allDomainsFile;
	$outPrefix =~ s/\.hmmhits.*$//i;
	dprint( "starting reduceDomains - domains\n", $opts );
	my $cmd = "$baseDir/bin/reduceDomains.pl -domains $allDomainsFile -db $opts->{f} -reduce domains -numThreads $opts->{j} -prefix $outPrefix";
	my $optParams = Util::configParams( $conf, "CDHIT_domains" );
	$cmd .= " $optParams"
		if ( length($optParams) > 0 );
	system( $cmd );

	dprint( "starting reduceDomains - others\n", $opts );
	$cmd = "$baseDir/bin/reduceDomains.pl -domains $allDomainsFile -db $opts->{f} -reduce other -numThreads $opts->{j} -prefix $outPrefix";
	$optParams = Util::configParams( $conf, "CDHIT_other" );
	$cmd .= " $optParams"
		if ( length($optParams) > 0 );
	system( $cmd );
	dprint( "second pass cd-hit reduce - others\n", $opts );
	$cmd = "$baseDir/bin/cd-hit -i $outPrefix.reduce.other.faa -o $outPrefix.reduce.other.cdhit.faa";
	$optParams = Util::configParams( $conf, "CDHIT_other2" );
	$cmd .= " $optParams"
		if ( length($optParams) > 0 );
	system( $cmd );
	dprint( "cleaning up cd-hit\n", $opts );
	unlink( "$outPrefix.reduce.other.cdhit.faa.bak.clstr" )
		if ( !exists( $opts->{D} ) );
	rename( "$outPrefix.reduce.other.cdhit.faa.clstr", "$outPrefix.reduce.other.cdhit.clstr" );

	dprint( "building combined and parsed clusters file\n", $opts );
	$cmd = "$baseDir/bin/clstr_rev.pl $outPrefix.reduce.other.clstr $outPrefix.reduce.other.cdhit.clstr >$outPrefix.reduce.other.comb.clstr";
	system( $cmd );
	unlink( "$outPrefix.reduce.other.clstr", "$outPrefix.reduce.other.cdhit.clstr" )
		if ( !exists( $opts->{D} ) );
	concatenateFiles( "$opts->{o}/combined.clstr", "$outPrefix.reduce.domains.clstr", "$outPrefix.reduce.other.comb.clstr" );
	unlink( "$outPrefix.reduce.domains.clstr", "$outPrefix.reduce.other.comb.clstr" )
		if ( !exists( $opts->{D} ) );
	generateParsedClusters( "$opts->{o}/combined.clstr", "$opts->{o}/combined.clstr.parsed" );
	unlink( "$opts->{o}/combined.clstr" )
		if ( !exists( $opts->{D} ) );
	$cmd = "$baseDir/bin/sort -u < $opts->{o}/combined.clstr.parsed | $baseDir/bin/sort -n > $opts->{o}/combined.clstr.parsed.sorted";
	system( $cmd );
	unlink( "$opts->{o}/combined.clstr.parsed" )
		if ( !exists( $opts->{D} ) );

	dprint( "merging redundant domains between families\n", $opts );
	$cmd = "$baseDir/bin/mergeDomains.pl -sort -domains $outPrefix.reduce.domains -db $outPrefix.reduce.domains.faa";
	$optParams = Util::configParams( $conf, "MERGEDOMAINS" );
	$cmd .= " $optParams"
		if ( length($optParams) > 0 );
	system( $cmd );

	dprint( "concatenating merged domains and other domains faa files\n", $opts );
	concatenateFiles( "$opts->{o}/combined.faa", "$outPrefix.reduce.domains.merged.faa", "$outPrefix.reduce.other.cdhit.faa" );
	$cmd = "$baseDir/bin/formatdb -p T -i $opts->{o}/combined.faa";
	system( $cmd );

	if ( exists( $opts->{b} ) )
	{
		dprint( "splitting blast query files for distributed computation (b = $opts->{b})\n", $opts );
		generateSplitFiles( "$outPrefix.reduce.other.cdhit.faa", $opts );
	} else {
		local *OUT;
		open(OUT, ">$opts->{o}/file.list");
		print OUT "$outPrefix.reduce.other.cdhit.faa\n";
		close(OUT);
	}
}

sub generateParsedClusters
{
	my $inFile = shift;
	my $outFile = shift;
	local *IN;
	local *OUT;

	open(IN, "<$inFile");
	open(OUT, ">$outFile");

	my $outLine = undef;
	my @before = ();
	my $exemplar = undef;
	while(<IN>)
	{
		if ( /^>/ )
		{
			print OUT $outLine, "\n"
				if ( defined($outLine) );
			$outLine = "";
			@before = ();
			$exemplar = undef;
		} elsif ( /^\d+\s+\d+aa,\s+>(\S+\.\d+\.\d+)\.+\s+(.+)/ )
		{
			my $seqId = $1;
			my $star = $2;
			if ( $star eq '*' )
			{
				$outLine = "$seqId\t$seqId";
				if ( scalar( @before ) > 0 )
				{
					$outLine .= "\t" . join( "\t", @before );
					@before = ();
				}
				$exemplar = $seqId;
			} else {
				if ( defined($exemplar) )
				{
					$outLine .= "\t$seqId";
				} else {
					push( @before, $seqId );
				}
			}
		}
	}
	print OUT $outLine, "\n";

	close(OUT);
	close(IN);
}

sub generateSplitFiles
{
	my $inFile = shift;
	my $opts = shift;
	my $batchSize = $opts->{b};
	my $outDir = $opts->{o};

	my $batchNum = 0;
	my $batchCount = 0;
	local *IN;
	local *OUT;
	local *FL;

	open(FL, ">$outDir/file.list");
	open(IN, "<$inFile");
	open(OUT, ">$inFile.$batchNum");

	while( my $line = <IN> )
	{
		if ( $line =~ /^>/ )
		{
			if ( $batchCount >= $batchSize )
			{
				print FL "$inFile.$batchNum\n";
				close(OUT);
				$batchNum++;
				open(OUT, ">$inFile.$batchNum");
				$batchCount = 1;
			} else {
				$batchCount++;
			}
		}

		print OUT $line;
	}
	print FL "$inFile.$batchNum\n";

	close(OUT);
	close(IN);
	close(FL);
}

sub concatenateFiles
{
	my $outFile = shift;
	local *OUT;

	open(OUT, ">>$outFile");

	while ( my $file = shift )
	{
		local *IN;
		open(IN, "<$file");
		while(<IN>)
		{
			print OUT;
		}
		close(IN);
	}

	close(OUT);
}

sub loadPfamClanMap
{
	my $opts = shift;
	my %map = ();
	local *IN;

	my $mapFile = $opts->{d} . "/pfam/Pfam-C.map";
	if ( !(-e $mapFile) )
	{
		dprint( "[warning] pfam id to clan map does not exist in '$opts->{d}'\n", $opts );
		dprint( "[warning] unable to map pfam ids to clan ids - performance may be worse!\n", $opts );
		return \%map;
	}

	open(IN, "<$mapFile");
	while(<IN>)
	{
		my ( $clanId, $pfamId ) = split(/\s+/, $_);
		$map{$pfamId} = $clanId;
	}
	close(IN);

	return \%map;
}

sub mapPfamClan
{
	my $pfamId = shift;
	my $clanMap = shift;

	return $pfamId
		if ( !defined($clanMap) || !exists($clanMap->{$pfamId}) );

	return $clanMap->{$pfamId};
}

sub concatenateFastHmmDomains
{
	my $files = shift;
	my $opts = shift;

	my ( $baseFile ) = (split(/\//, $files->[0]))[-1];
	my @fileParts = split(/\./, $baseFile, 4);
	my $outFile = $opts->{o} . "/result." . $fileParts[1] . ".all.hmmhits";
	local *IN;
	local *OUT;

	open(OUT, ">$outFile");
	foreach my $file ( @{$files} )
	{
		my ( $baseFile ) = (split(/\//, $file))[-1];
		my ( undef, undef, $dbName ) = split(/\./, $baseFile);
		if ( $dbName ne 'pfam' )
		{
			open(IN, "<$file");
			while(<IN>)
			{
				print OUT;
			}
			close(IN);
		} else {
			# map pfam id to clan id if pfam is in a clan
			my $clanMap = loadPfamClanMap( $opts );
			open(IN, "<$file");
			while(<IN>)
			{
				my @pfamCols = split(/\t/, $_);
				$pfamCols[0] = mapPfamClan( $pfamCols[0], $clanMap );
				print OUT join("\t", @pfamCols);
			}
			close(IN);
		}
	}
	close(OUT);

	return $outFile;
}

sub dprint
{
	my $msg = shift;
	my $opts = shift;
	print "[" . localtime() . "] $msg"
		if ( !exists( $opts->{q} ) );
}
