#!/usr/bin/perl -w
#
#  $Id: mergeDomains.pl,v 1.2 2007/10/16 23:20:20 whuang Exp $
#  fastHmm/fastBlast Alignment Tools
#  http://microbesonline.org/fasthmm (fasthmm@microbesonline.org)
#
#  Merge domain families into representative set
#
#  Copyright (C) 2007 The Regents of the University of California
#  All rights reserved.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#  Disclaimer
#
#  NEITHER THE UNITED STATES NOR THE UNITED STATES DEPARTMENT OF ENERGY,
#  NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED,
#  OR ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY,
#  COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
#  OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
#  PRIVATELY OWNED RIGHTS.
#

use strict;

use lib exists( $ENV{FASTHMM_DIR} ) ?
		"$ENV{FASTHMM_DIR}/lib" :
		"./lib";
use Args;
use File;
use Util;
use Vector;


#
# USAGE
#
my $usage =
"Usage:
  mergeDomains.pl <options>

Parameters:
  -domains <inFile>	Specify reduced domains file
  -db      <blastDb>	Specify corresponding extracted subsequence FASTA file

Optional Parameters:
  -sort			Sort domains (required if domains are unsorted)
  -stats <n>		Display stats for every <n> domains parsed
  -prefix <outFile>	Specify output files prefix;
			Two files will be created <outFile>.domains and
			<outFile>.faa containing the non-overlapping
			merged domains and their corresponding extracted
			subsequence FASTA file.  If <outFile> is '-'
			only the extracted subsequence FASTA file will be
			output to STDOUT.
			Default: <inFile>.merged

";

#
# Parse Command Line Options
#
my ($opts, $nonOpts) = Args::getArgs( 
	"+domains:|+db:|prefix:|stats:|sort",
	@ARGV, -1, $usage );

die "Error: Specified input file '", $opts->{domains}, "' does not exist!\n"
	if ( !(-e $opts->{domains}) );

die "Error: Specified extracted sequence file '", $opts->{db}, "' does not exist!\n"
	if ( !(-e $opts->{db}) );

$opts->{prefix} = $opts->{domains} . ".merged"
	if ( !exists( $opts->{prefix} ) );


mergeDomains( $opts );

exit(0);


sub mergeDomains
{
	my $opts = shift;

	# open the sequence file using fasta iterator
	local *IN;
	# the following enables debugging output (cache sizes, hits, etc)
	#my $fId = File::openFastaCache( \*IN, $opts->{db}, undef, undef, 0, 1 );
	my $fId = File::openFastaCache( \*IN, $opts->{db} );

	local *DOUT;
	local *FOUT;

	if ( $opts->{prefix} eq '-' )
	{
		open(FOUT, ">&=1");
	} else {
		open(DOUT, ">$opts->{prefix}.domains");
		open(FOUT, ">$opts->{prefix}.faa");
	}

	# open and iterate over the input domain file
	my $curLocusId = undef;
	my $curRegionBeg = -1;
	my $curRegionEnd = -1;
	my $curRegionSeq = "";

	# stats
	my $numDomains = 0;
	my $numMerged = 0;
	my $lenDomains = 0;
	my $lenMerged = 0;

	local *DIN;
	if ( exists( $opts->{sort} ) )
	{
		my $baseDir = exists( $ENV{FASTHMM_DIR} ) ?
				$ENV{FASTHMM_DIR} :
				".";
		open(DIN, "$baseDir/bin/sort -nk2 -nk3 -nk4 -u $opts->{domains} |");
	} else {
		open(DIN, "<$opts->{domains}");
	}
	while(<DIN>)
	{
		s/^\s+|\s+$//g;
		my ( undef, $locusId, $locBeg, $locEnd ) = split(/\t/, $_, 4);
		my $seqId = join(".", $locusId, $locBeg, $locEnd);
		my $seq = File::getFastaCacheSeq( $fId, $seqId );

		$numDomains++;
		$lenDomains += length($seq);

		if ( !defined($curLocusId) ||
			( $locusId ne $curLocusId ) )
		{
			# the current region is it for $curLocusId
			if ( defined($curLocusId) )
			{
				my $regSeqId = join(".", $curLocusId, $curRegionBeg, $curRegionEnd);
				print FOUT ">$regSeqId\n$curRegionSeq\n";
				print DOUT join("\t", $regSeqId, $regSeqId, $curRegionBeg, $curRegionEnd) . "\n"
					if ( $opts->{prefix} ne '-' );

				$numMerged++;
				$lenMerged += length($curRegionSeq);
			}
			$curLocusId = $locusId;
			$curRegionBeg = $locBeg;
			$curRegionEnd = $locEnd;
			$curRegionSeq = $seq;
		} else {
			if ( $locBeg <= $curRegionEnd + 1 )
			{
				# flanking
				if ( $locEnd > $curRegionEnd )
				{
					# merge current region with local region
					$curRegionSeq .= substr( $seq, $curRegionEnd - $locBeg );
					$curRegionEnd = $locEnd;
				} else {
					# contained - ignore
				}
			} else {
				my $regSeqId = join(".", $curLocusId, $curRegionBeg, $curRegionEnd);
				print FOUT ">$regSeqId\n$curRegionSeq\n";
				print DOUT join("\t", $regSeqId, $regSeqId, $curRegionBeg, $curRegionEnd) . "\n"
					if ( $opts->{prefix} ne '-' );

				$numMerged++;
				$lenMerged += length($curRegionSeq);

				$curRegionBeg = $locBeg;
				$curRegionEnd = $locEnd;
				$curRegionSeq = $seq;
			}
		}

		if ( exists( $opts->{stats} ) )
		{
			if ( $numDomains % $opts->{stats} == 0 )
			{
				print STDERR sprintf( " ... %s domains (%s) merged to %s domains (%s): %.2f%% bp\n",
					Util::commify($numDomains),
					Util::formatBp($lenDomains),
					Util::commify($numMerged),
					Util::formatBp($lenMerged),
					( 100.0 * $lenMerged / $lenDomains ) );
			}
		}
	}
	close(DIN);

	# left-overs
	my $regSeqId = join(".", $curLocusId, $curRegionBeg, $curRegionEnd);
	print FOUT ">$regSeqId\n$curRegionSeq\n";
	print DOUT join("\t", $regSeqId, $regSeqId, $curRegionBeg, $curRegionEnd) . "\n"
		if ( $opts->{prefix} ne '-' );

	$numMerged++;
	$lenMerged += length($curRegionSeq);

	# final stats (always display?)
	print STDERR sprintf( "Done: %s domains (%s) merged to %s domains (%s): %.2f%% bp\n",
					Util::commify($numDomains),
					Util::formatBp($lenDomains),
					Util::commify($numMerged),
					Util::formatBp($lenMerged),
					( 100.0 * $lenMerged / $lenDomains ) );

	# close fasta iterator
	File::closeFastaCache( $fId );

	# close output
	close(DOUT);
	close(FOUT);
}
